@@ -3,8 +3,9 @@ from pysnmp.proto.mpmod.base import AbstractMessageProcessingModel
 from pysnmp.proto.secmod import rfc3414
 from pysnmp.proto import rfc1905, rfc3411, error, api
 from pyasn1.type import univ, namedtype, constraint
-from pyasn1.codec.ber import encoder, decoder
+from pyasn1.codec.ber import decoder
 from pyasn1.error import PyAsn1Error
+from pysnmp import debug
 
 # API to rfc1905 protocol objects
 pMod = api.protoModules[api.protoVersion2c]
@@ -56,6 +57,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
     def __init__(self):
         AbstractMessageProcessingModel.__init__(self)
         self.__engineIDs = {}
+        self.__engineIDsExpQueue = {}
+        self.__expirationTimer = 0L
         
     # 7.1.1a
     def prepareOutgoingMessage(
@@ -80,6 +83,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         # 7.1.1b
         msgID = self._newMsgID()
 
+        debug.logger & debug.flagMP and debug.logger('prepareOutgoingMessage: new msgID %s' % msgID)
+        
         peerSnmpEngineData = self.__engineIDs.get(
             (transportDomain, transportAddress)
             )
@@ -95,11 +100,12 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                 # is a sort of workaround.
                 if not contextEngineId:
                     contextEngineId = peerSnmpEngineData['securityEngineID']
-
         # 7.1.5
         if not contextName:
             contextName = ''
 
+        debug.logger & debug.flagMP and debug.logger('prepareOutgoingMessage: using contextEngineId %s, contextName %s' % (contextEngineId, contextName))
+        
         # 7.1.6
         scopedPDU = ScopedPDU()
         scopedPDU.setComponentByPosition(0, contextEngineId)
@@ -123,7 +129,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         snmpEngineMaxMessageSize, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineMaxMessageSize')
 
         # 7.1.7c
-        headerData.setComponentByPosition(1, snmpEngineMaxMessageSize.syntax)
+        # XXX need to coerce MIB value as it has incompatible constraints set
+        headerData.setComponentByPosition(1, int(snmpEngineMaxMessageSize.syntax))
 
         # 7.1.7d
         msgFlags = 0
@@ -144,7 +151,10 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         headerData.setComponentByPosition(2, chr(msgFlags))
 
         # 7.1.7e
-        headerData.setComponentByPosition(3, securityModel)
+        # XXX need to coerce MIB value as it has incompatible constraints set
+        headerData.setComponentByPosition(3, int(securityModel))
+
+        debug.logger & debug.flagMP and debug.logger('prepareOutgoingMessage: %s' % (msg.prettyPrint(),))
 
         smHandler = snmpEngine.securityModels.get(securityModel)
         if smHandler is None:
@@ -175,9 +185,12 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                 scopedPDU.getComponentByPosition(2).setComponentByType(
                     emptyPdu.tagSet, emptyPdu
                 )
+                debug.logger & debug.flagMP and debug.logger('prepareOutgoingMessage: force engineID discovery')
             else:
                 securityEngineID = peerSnmpEngineData['securityEngineID']
-                
+
+        debug.logger & debug.flagMP and debug.logger('prepareOutgoingMessage: securityEngineID %s' % securityEngineID)
+             
         # 7.1.9.b
         ( securityParameters,
           wholeMsg ) = smHandler.generateRequestMsg(
@@ -248,7 +261,9 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         maxMessageSize = cachedParams['msgMaxSize']
         transportDomain = cachedParams['transportDomain']
         transportAddress = cachedParams['transportAddress']
-            
+
+        debug.logger & debug.flagMP and debug.logger('prepareResponseMessage: stateReference %s' % (stateReference))
+        
         # 7.1.3
         if statusInformation is not None and statusInformation.has_key('oid'):
             # 7.1.3a
@@ -300,6 +315,7 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
             # 7.1.3e
             pdu = reportPDU
 
+            debug.logger & debug.flagMP and debug.logger('prepareResponseMessage: prepare report PDU for statusInformation %s' % statusInformation)
         # 7.1.4
         if not contextEngineId:
             contextEngineId = snmpEngineID  # XXX impl-dep manner
@@ -308,6 +324,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         if not contextName:
             contextName = ''
 
+        debug.logger & debug.flagMP and debug.logger('prepareResponseMessage: using contextEngineId %s, contextName %s' % (contextEngineId, contextName))
+
         # 7.1.6
         scopedPDU = ScopedPDU()
         scopedPDU.setComponentByPosition(0, contextEngineId)
@@ -331,7 +349,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         snmpEngineMaxMessageSize, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineMaxMessageSize')
 
         # 7.1.7c
-        headerData.setComponentByPosition(1, snmpEngineMaxMessageSize.syntax)
+        # XXX need to coerce MIB value as it has incompatible constraints set
+        headerData.setComponentByPosition(1, int(snmpEngineMaxMessageSize.syntax))
 
         # 7.1.7d
         msgFlags = 0
@@ -354,6 +373,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         # 7.1.7e
         headerData.setComponentByPosition(3, securityModel)
 
+        debug.logger & debug.flagMP and debug.logger('prepareResponseMessage: %s' % (msg.prettyPrint(),))
+
         smHandler = snmpEngine.securityModels.get(securityModel)
         if smHandler is None:
             raise error.StatusInformation(
@@ -379,6 +400,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
             # 7.1.8.b            
             raise
 
+        debug.logger & debug.flagMP and debug.logger('prepareResponseMessage: SM finished')
+
         # Message size constraint verification
         if len(wholeMsg) > min(snmpEngineMaxMessageSize.syntax, maxMessageSize):
             raise error.StatusInformation(errorIndication='tooBig')
@@ -406,6 +429,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                 errorIndication = 'parseError'
                 )
 
+        debug.logger & debug.flagMP and debug.logger('prepareDataElements: %s' % (msg.prettyPrint(),))
+
         # 7.2.3
         headerData = msg.getComponentByPosition(1)
         msgVersion = messageProcessingModel = msg.getComponentByPosition(0)
@@ -414,6 +439,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         maxMessageSize = headerData.getComponentByPosition(1)
         securityModel = headerData.getComponentByPosition(3)
         securityParameters = msg.getComponentByPosition(2)
+
+        debug.logger & debug.flagMP and debug.logger('prepareDataElements: msg data msgVersion %s msgID %s securityModel %s' % (msgVersion, msgID, securityModel))
         
         # 7.2.4
         if not snmpEngine.securityModels.has_key(securityModel):
@@ -459,7 +486,9 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                 wholeMsg,
                 msg
                 )
+            debug.logger & debug.flagMP and debug.logger('prepareDataElements: SM succeeded')
         except error.StatusInformation, statusInformation:
+            debug.logger & debug.flagMP and debug.logger('prepareDataElements: SM failed, statusInformation %s' % statusInformation)
             if statusInformation.has_key('errorIndication'):
                 # 7.2.6a
                 if statusInformation.has_key('oid'):
@@ -516,7 +545,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                             )
                     except error.StatusInformation:
                         pass
-    
+                    
+                    debug.logger & debug.flagMP and debug.logger('prepareDataElements: error reported')
             # 7.2.6b
             raise statusInformation
         else:
@@ -524,12 +554,20 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
             k = (transportDomain, transportAddress)
             if not self.__engineIDs.has_key(k):
                 contextEngineId, contextName, pdu = scopedPDU
+                
                 self.__engineIDs[k] = {
                     'securityEngineID': securityEngineID,
                     'contextEngineId': contextEngineId,
                     'contextName': contextName
                     }
 
+                expireAt = self.__expirationTimer + 300
+                if not self.__engineIDsExpQueue.has_key(expireAt):
+                    self.__engineIDsExpQueue[expireAt] = []
+                self.__engineIDsExpQueue[expireAt].append(k)
+                    
+                debug.logger & debug.flagMP and debug.logger('prepareDataElements: cache securityEngineID %s for %s %s' % (securityEngineID, transportDomain, transportAddress))
+
         snmpEngineID, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')
         snmpEngineID = snmpEngineID.syntax
 
@@ -558,6 +596,8 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         else:
             sendPduHandle = None
 
+        debug.logger & debug.flagMP and debug.logger('prepareDataElements: using sendPduHandle %s for msgID %s' % (sendPduHandle, msgID))
+        
         # 7.2.11
         if rfc3411.internalClassPDUs.has_key(pduType):
             # 7.2.11a
@@ -584,7 +624,7 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
             raise statusInformation
 
         statusInformation = None  # no errors ahead
-        
+
         # 7.2.12
         if rfc3411.responseClassPDUs.has_key(pduType):
             # 7.2.12a -> noop
@@ -644,7 +684,9 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
                 transportDomain=transportDomain,
                 transportAddress=transportAddress
                 )            
-            
+
+            debug.logger & debug.flagMP and debug.logger('prepareDataElements: new stateReference %s' % stateReference)
+
             # 7.2.13c
             return ( messageProcessingModel,
                      securityModel,
@@ -679,3 +721,14 @@ class SnmpV3MessageProcessingModel(AbstractMessageProcessingModel):
         raise error.StatusInformation(
             errorIndication = 'unknownPDU'
             )
+
+    def __expireEnginesInfo(self):
+        if self.__engineIDsExpQueue.has_key(self.__expirationTimer):
+            for engineKey in self.__engineIDsExpQueue[self.__expirationTimer]:
+                del self.__engineIDs[engineKey]
+                debug.logger & debug.flagMP and debug.logger('__expireEnginesInfo: expiring %s' % (engineKey,))
+            del self.__engineIDsExpQueue[self.__expirationTimer]
+        self.__expirationTimer = self.__expirationTimer + 1
+        
+    def receiveTimerTick(self, snmpEngine, timeNow):
+        self.__expireEnginesInfo()
