@@ -6,6 +6,7 @@ from pysnmp.proto.api import verdec # XXX
 from pysnmp.carrier.asynsock.dispatch import AsynsockDispatcher
 from pysnmp.carrier.asynsock.dgram import udp
 from pysnmp.error import PySnmpError
+from pysnmp import debug
 
 class MsgAndPduDispatcher:
     """SNMP engine PDU & message dispatcher. Exchanges SNMP PDU's with
@@ -74,7 +75,8 @@ class MsgAndPduDispatcher:
 
             # 4.3.4
             self.__appsRegistration[k] = processPdu
-        
+
+        debug.logger & debug.flagDsp and debug.logger('registerContextEngineId: contextEngineId %s pduTypes %s' % (contextEngineId, pduTypes))
     # 4.4.1
     def unregisterContextEngineId(self, contextEngineId, pduTypes):
         """Unregister application with dispatcher"""
@@ -88,6 +90,8 @@ class MsgAndPduDispatcher:
             if self.__appsRegistration.has_key(k):
                 del self.__appsRegistration[k]
 
+        debug.logger & debug.flagDsp and debug.logger('unregisterContextEngineId: contextEngineId %s pduTypes %s' % (contextEngineId, pduTypes))
+
     def getRegisteredApp(self, contextEngineId, pduType):
         k = ( str(contextEngineId), pduType )
         if self.__appsRegistration.has_key(k):
@@ -116,8 +120,6 @@ class MsgAndPduDispatcher:
         expectResponse
         ):
         """PDU dispatcher -- prepare and serialize a request or notification"""
-#        print 'sendPdu', PDU
-#        print transportDomain, transportAddress, messageProcessingModel, securityModel, securityName, securityLevel, contextEngineId, contextName, pduVersion
         # 4.1.1.2
         mpHandler = snmpEngine.messageProcessingSubsystems.get(
             int(messageProcessingModel)
@@ -127,13 +129,20 @@ class MsgAndPduDispatcher:
                 errorIndication='unsupportedMsgProcessingModel'
                 )
 
+        debug.logger & debug.flagDsp and debug.logger('sendPdu: PDU %s' % PDU.prettyPrint())
+
         # 4.1.1.3
         sendPduHandle = self.__newSendPduHandle()
         if expectResponse:
             self.__cacheAdd(
-                sendPduHandle, expectResponse=expectResponse
+                sendPduHandle,
+                messageProcessingModel=messageProcessingModel,
+                sendPduHandle=sendPduHandle,
+                expectResponse=expectResponse
                 )
 
+        debug.logger & debug.flagDsp and debug.logger('sendPdu: new sendPduHandle %s' % sendPduHandle)
+
         # 4.1.1.4 & 4.1.1.5
         try:
             ( destTransportDomain,
@@ -153,8 +162,10 @@ class MsgAndPduDispatcher:
                 expectResponse,
                 sendPduHandle
                 )
+            debug.logger & debug.flagDsp and debug.logger('sendPdu: MP succeeded')
         except error.StatusInformation, statusInformation:
-#            self.releaseStateInformation(snmpEngine, sendPduHandle)
+# XXX is it still needed here?
+#            self.releaseStateInformation(snmpEngine, sendPduHandle, messageProcessingModel)
             raise
 
         # 4.1.1.6
@@ -170,16 +181,13 @@ class MsgAndPduDispatcher:
                 sendPduHandle,
                 transportDomain=transportDomain,
                 transportAddress=transportAddress,
-                messageProcessingModel=messageProcessingModel,
                 securityModel=securityModel,
                 securityName=securityName,
                 securityLevel=securityLevel,
                 contextEngineId=contextEngineId,
                 contextName=contextName,
                 pduVersion=pduVersion,
-                PDU=PDU,
-                expectResponse=expectResponse,
-                sendPduHandle=sendPduHandle,
+                PDU=PDU
                 )
 
         return sendPduHandle
@@ -201,7 +209,6 @@ class MsgAndPduDispatcher:
         statusInformation
         ):
         """PDU dispatcher -- prepare and serialize a response"""
-#        print 'returnResponsePdu', PDU, statusInformation
         # Extract input values and initialize defaults
         mpHandler = snmpEngine.messageProcessingSubsystems.get(
             int(messageProcessingModel)
@@ -211,6 +218,8 @@ class MsgAndPduDispatcher:
                 errorIndication='unsupportedMsgProcessingModel'
                 )
 
+        debug.logger & debug.flagDsp and debug.logger('returnResponsePdu: PDU %s' % (PDU and PDU.prettyPrint() or "<empty>",))
+
         # 4.1.2.2
         try:
             ( destTransportDomain,
@@ -229,6 +238,7 @@ class MsgAndPduDispatcher:
                 stateReference,
                 statusInformation
                 )
+            debug.logger & debug.flagDsp and debug.logger('returnResponsePdu: MP suceeded')
         except error.StatusInformation, statusInformation:
             # 4.1.2.3
             raise
@@ -257,7 +267,6 @@ class MsgAndPduDispatcher:
         wholeMsg
         ):
         """Message dispatcher -- de-serialize message into PDU"""
-#        print 'receiveMessage', time.time() #, repr(wholeMsg)
         # 4.2.1.1
         snmpInPkts, = self.mibInstrumController.mibBuilder.importSymbols(
             '__SNMPv2-MIB', 'snmpInPkts'
@@ -273,15 +282,15 @@ class MsgAndPduDispatcher:
             snmpInAsn1ParseErrs.syntax = snmpInAsn1ParseErrs.syntax + 1
             return ''  # n.b the whole buffer gets dropped
 
+        debug.logger & debug.flagDsp and debug.logger('receiveMessage: msgVersion %s, msg decoded' % msgVersion)
+
         messageProcessingModel = msgVersion
         
         mpHandler = snmpEngine.messageProcessingSubsystems.get(
             int(messageProcessingModel)
             )
         if mpHandler is None:
-            snmpInBadVersions, = self.mibInstrumController.mibBuilder.importSymbols(
-                '__SNMPv2-MIB', 'snmpInBadVersions'
-                )
+            snmpInBadVersions, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInBadVersions')
             snmpInBadVersions.syntax = snmpInBadVersions.syntax + 1
             return restOfWholeMsg
 
@@ -307,10 +316,12 @@ class MsgAndPduDispatcher:
                 transportAddress,
                 wholeMsg
                 )
+            debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP succeded')
         except error.StatusInformation, statusInformation:
             if statusInformation.has_key('sendPduHandle'):
                 # Dropped REPORT -- re-run pending reqs queue as some
                 # of them may be waiting for this REPORT
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP failed, statusInformation %s' % statusInformation)
                 self.__expireRequest(
                     snmpEngine,
                     self.__cachePop(statusInformation['sendPduHandle']),
@@ -318,14 +329,16 @@ class MsgAndPduDispatcher:
                     )
             return restOfWholeMsg
 
-#        print 'recv', PDU
+        debug.logger & debug.flagDsp and debug.logger('receiveMessage: PDU %s' % PDU.prettyPrint())
+
         # 4.2.2
         if sendPduHandle is None:
             # 4.2.2.1 (request or notification)
 
+            debug.logger & debug.flagDsp and debug.logger('receiveMessage: pduType %s' % pduType)
             # 4.2.2.1.1
             processPdu = self.getRegisteredApp(contextEngineId, pduType)
-
+            
             # 4.2.2.1.2
             if processPdu is None:
                 # 4.2.2.1.2.a
@@ -339,6 +352,8 @@ class MsgAndPduDispatcher:
                     'val': snmpUnknownPDUHandlers.syntax
                     }                    
 
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: unhandled PDU type')
+                
                 # XXX fails on unknown PDU
                 
                 try:
@@ -359,6 +374,7 @@ class MsgAndPduDispatcher:
                         statusInformation
                         )
                 except error.StatusInformation, statusInformation:
+                    debug.logger & debug.flagDsp and debug.logger('receiveMessage: report failed, statusInformation %s' % statusInformation)
                     return restOfWholeMsg
                 
                 # 4.2.2.1.2.c
@@ -371,6 +387,8 @@ class MsgAndPduDispatcher:
                 except PySnmpError: # XXX
                     pass
 
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: reporting succeeded')
+                
                 # 4.2.2.1.2.d
                 return restOfWholeMsg
             else:
@@ -388,6 +406,7 @@ class MsgAndPduDispatcher:
                     maxSizeResponseScopedPDU,
                     stateReference
                     )
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: processPdu succeeded')
                 return restOfWholeMsg
         else:
             # 4.2.2.2 (response)
@@ -401,6 +420,8 @@ class MsgAndPduDispatcher:
                 snmpUnknownPDUHandlers.syntax = snmpUnknownPDUHandlers.syntax+1
                 return restOfWholeMsg
 
+            debug.logger & debug.flagDsp and debug.logger('receiveMessage: cache read by sendPduHandle %s' % sendPduHandle)
+            
             # 4.2.2.2.3
             # no-op ? XXX
 
@@ -422,7 +443,7 @@ class MsgAndPduDispatcher:
                 cachedParams['sendPduHandle'],
                 cbCtx
                 )
-
+            debug.logger & debug.flagDsp and debug.logger('receiveMessage: processResponsePdu succeeded')
             return restOfWholeMsg
 
     def releaseStateInformation(
@@ -439,6 +460,9 @@ class MsgAndPduDispatcher:
         processResponsePdu, timeoutAt, cbCtx = cachedParams['expectResponse']
         if statusInformation is None and time.time() < timeoutAt:
             return
+
+        debug.logger & debug.flagDsp and debug.logger('__expireRequest: req cachedParams %s' % cachedParams)
+
         # Fail timed-out requests        
         if not statusInformation:
             statusInformation = error.StatusInformation(
