@@ -8,6 +8,8 @@ from pysnmp.proto import rfc1155, error
 from pyasn1.type import univ, namedtype, constraint
 from pyasn1.codec.ber import encoder, decoder
 from pyasn1.error import PyAsn1Error
+from pysnmp import debug
+import time
 
 # USM security params
 
@@ -37,6 +39,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
     def __init__(self):
         AbstractSecurityModel.__init__(self)
         self.__timeline = {}
+        self.__timelineExpQueue = {}
+        self.__expirationTimer = 0L
 
     def __getUserInfo(
         self, mibInstrumController, securityEngineID, securityName
@@ -208,6 +212,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 'usmUserPrivKeyLocalized'
                 )
             securityEngineID = snmpEngineID
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: user info read from cache')
         elif securityName:
             # 3.1.1b
             try:
@@ -219,6 +224,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     snmpEngine.msgAndPduDsp.mibInstrumController,
                     securityEngineID, securityName
                     )
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: read user info')
             except NoSuchInstanceError:
                 pysnmpUsmDiscovery, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__PYSNMP-USM-MIB', 'pysnmpUsmDiscovery')
                 __reportUnknownName = not pysnmpUsmDiscovery.syntax
@@ -235,15 +241,20 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                             )
                     except NoSuchInstanceError:
                         __reportUnknownName = 1
+
                 if __reportUnknownName:
                     raise error.StatusInformation(
                         errorIndication = 'unknownSecurityName'
                         )
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: clone user info')
         else:
             # empty username used for engineID discovery
             usmUserName = usmUserSecurityName = ''
             usmUserAuthProtocol = usmUserAuthKeyLocalized = None
             usmUserPrivProtocol = usmUserPrivKeyLocalized = None
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use empty USM data')
+            
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: local user usmUserName %s usmUserAuthProtocol %s usmUserPrivProtocol %s by securityEngineID %s securityName %s' % (usmUserName, usmUserAuthProtocol, usmUserPrivProtocol, securityEngineID, securityName))
 
         msg = globalData
         
@@ -274,6 +285,9 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     errorIndication = 'encryptionError'
                     )
             dataToEncrypt = encoder.encode(scopedPDU)
+            
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU encoded')
+            
             try:
                 ( encryptedData,
                   privParameters ) = privHandler.encryptData(
@@ -283,6 +297,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             except error.StatusInformation, statusInformation:
                 raise
 
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU cyphered')
+
             securityParameters.setComponentByPosition(5, privParameters)
             scopedPDUData.setComponentByPosition(1, encryptedData)
 
@@ -300,22 +316,29 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             if self.__timeline.has_key(securityEngineID):
                 ( snmpEngineBoots,
                   snmpEngineTime,
-                  latestReceivedEngineTime ) = self.__timeline[
+                  latestReceivedEngineTime,
+                  latestUpdateTimestamp) = self.__timeline[
                     securityEngineID
                     ]
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: read snmpEngineBoots, snmpEngineTime from timeline')
             else:
                 # 2.3 XXX is this correct?
                 snmpEngineBoots = snmpEngineTime = 0
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: no timeline for securityEngineID %s' % securityEngineID)
         # 3.1.6.b
         elif securityStateReference is not None:  # XXX Report?
             ( snmpEngineBoots,
               snmpEngineTime ) = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineBoots', 'snmpEngineTime')
             snmpEngineBoots = snmpEngineBoots.syntax
-            snmpEngineTime = snmpEngineTime.syntax
+            snmpEngineTime = snmpEngineTime.syntax.clone()
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: read snmpEngineBoots, snmpEngineTime from LCD')
         # 3.1.6.c
         else:
             snmpEngineBoots = snmpEngineTime = 0
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: assuming zero snmpEngineBoots, snmpEngineTime')
 
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use snmpEngineBoots %s snmpEngineTime %s for securityEngineID %s' % (snmpEngineBoots, snmpEngineTime, securityEngineID))
+        
         securityParameters.setComponentByPosition(1, snmpEngineBoots)
         securityParameters.setComponentByPosition(2, snmpEngineTime)
     
@@ -335,6 +358,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 4, '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                 )
 
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s' % (securityParameters.prettyPrint(),))
+            
             msg.setComponentByPosition(2, encoder.encode(securityParameters))
 
             wholeMsg = encoder.encode(msg)
@@ -346,11 +371,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             except error.StatusInformation, statusInformation:
                 raise
             
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: auth outgoing msg')
         # 3.1.8b
         else:
             securityParameters.setComponentByPosition(4, '')
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s' % (securityParameters.prettyPrint(),))
             msg.setComponentByPosition(2, encoder.encode(securityParameters))
             authenticatedWholeMsg = encoder.encode(msg)
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: plain outgoing msg')
 
         # 3.1.9
         return (
@@ -434,6 +462,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                errorIndication='parseError'
                )
 
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: %s' % (securityParameters.prettyPrint(),))
+
         # 3.2.9 -- moved up here to be able to report
         # maxSizeResponseScopedPDU on error
         maxSizeResponseScopedPDU = maxMessageSize - 512   # XXX
@@ -446,36 +476,49 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             msgUserName=securityParameters.getComponentByPosition(3)
             )
 
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: cache read securityStateReference %s by msgUserName %s' % (securityStateReference, securityParameters.getComponentByPosition(3)))
+        
         # Used for error reporting
         contextEngineId = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineID')[0].syntax
         contextName = ''
 
         # 3.2.3
         if not self.__timeline.has_key(securityEngineID):
+            debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %s' % securityEngineID)
             if securityEngineID:
                 # 3.2.3a XXX any other way to get auth engine in cache?
                 self.__timeline[securityEngineID] = (
                     securityParameters.getComponentByPosition(1),
                     securityParameters.getComponentByPosition(2),
-                    securityParameters.getComponentByPosition(2)
+                    securityParameters.getComponentByPosition(2),
+                    int(time.time())
                     )
+                
+                expireAt = self.__expirationTimer + 300
+                if not self.__timelineExpQueue.has_key(expireAt):
+                    self.__timelineExpQueue[expireAt] = []
+                self.__timelineExpQueue[expireAt].append(securityEngineID)
+                    
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: store timeline for securityEngineID %s' % (securityEngineID,))
             else:
                 # 3.2.3b
                 usmStatsUnknownEngineIDs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownEngineIDs')
                 usmStatsUnknownEngineIDs.syntax = usmStatsUnknownEngineIDs.syntax+1
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: null securityEngineID')
                 pysnmpUsmDiscoverable, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__PYSNMP-USM-MIB', 'pysnmpUsmDiscoverable')
                 if pysnmpUsmDiscoverable.syntax:
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: request EngineID discovery')
                     raise error.StatusInformation(
                         errorIndication = 'unknownEngineID',
                         oid=usmStatsUnknownEngineIDs.name,
                         val=usmStatsUnknownEngineIDs.syntax,
                         securityStateReference=securityStateReference,
-                        securityLevel=securityLevel,
                         contextEngineId=contextEngineId,
                         contextName=contextName,
                         maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
                         )
                 else:
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: will not discover EngineID')                    
                     # free securityStateReference XXX
                     raise error.StatusInformation(
                         errorIndication = 'unknownEngineID'
@@ -486,6 +529,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         msgAuthoritativeEngineID = securityParameters.getComponentByPosition(0)
         msgUserName = securityParameters.getComponentByPosition(3)
 
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read from securityParams msgAuthoritativeEngineID %s msgUserName %s' % (msgAuthoritativeEngineID, msgUserName))
+        
         if msgUserName:
             # 3.2.4
             try:
@@ -496,6 +541,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                   usmUserPrivKeyLocalized ) = self.__getUserInfo(
                     snmpEngine.msgAndPduDsp.mibInstrumController, msgAuthoritativeEngineID, msgUserName
                     )
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read user info from LCD')
             except NoSuchInstanceError:
                 pysnmpUsmDiscoverable, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__PYSNMP-USM-MIB', 'pysnmpUsmDiscoverable')
                 __reportUnknownName = not pysnmpUsmDiscoverable.syntax
@@ -510,8 +556,10 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                             securityEngineID,
                             msgUserName
                             )
+                        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: cloned user info')
                     except NoSuchInstanceError:
                         __reportUnknownName = 1
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %s msgUserName %s' % (securityEngineID, msgUserName))
                 if __reportUnknownName:
                         usmStatsUnknownUserNames, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownUserNames')
                         usmStatsUnknownUserNames.syntax = usmStatsUnknownUserNames.syntax+1
@@ -520,7 +568,6 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                             oid = usmStatsUnknownUserNames.name,
                             val = usmStatsUnknownUserNames.syntax,
                             securityStateReference=securityStateReference,
-                            securityLevel=securityLevel,
                             contextEngineId=contextEngineId,
                             contextName=contextName,
                             maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
@@ -531,6 +578,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             usmUserAuthProtocol = usmUserAuthKeyLocalized = None
             usmUserPrivProtocol = usmUserPrivKeyLocalized = None
 
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: now have usmUserSecurityName %s usmUserAuthProtocol %s usmUserPrivProtocol %s for msgUserName %s' % (usmUserSecurityName, usmUserAuthProtocol, usmUserPrivProtocol, msgUserName))
+
         # 3.2.5
         __reportError = 0
         if securityLevel == 3:
@@ -547,7 +596,6 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 oid=usmStatsUnknownEngineIDs.name,
                 val=usmStatsUnknownEngineIDs.syntax,
                 securityStateReference=securityStateReference,
-                securityLevel=securityLevel,
                 contextEngineId=contextEngineId,
                 contextName=contextName,
                 maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
@@ -574,22 +622,36 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     oid=usmStatsWrongDigests.name,
                     val=usmStatsWrongDigests.syntax,
                     securityStateReference=securityStateReference,
-                    securityLevel=securityLevel,
                     contextEngineId=contextEngineId,
                     contextName=contextName,
                     maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
                     )
-
+            
+            debug.logger & debug.flagSM and debug.logger('processIncomingMsg: incoming msg authed')
         # 3.2.7
         if securityLevel == 3 or securityLevel == 2:
-            if self.__timeline.has_key(securityEngineID):
+            if msgAuthoritativeEngineID == snmpEngineID:
+                # Authoritative SNMP engine: use local notion (SF bug #1649032)
                 ( snmpEngineBoots,
-                  snmpEngineTime,
-                  latestReceivedEngineTime ) = self.__timeline[
-                    msgAuthoritativeEngineID
-                    ]
+                  snmpEngineTime ) = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-FRAMEWORK-MIB', 'snmpEngineBoots', 'snmpEngineTime')
+                snmpEngineBoots = snmpEngineBoots.syntax
+                snmpEngineTime = snmpEngineTime.syntax.clone()
+                idleTime = 0
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read snmpEngineBoots (%s), snmpEngineTime (%s) from LCD' % (snmpEngineBoots, snmpEngineTime))
             else:
-                raise error.ProtocolError('Peer SNMP engine info missing')
+                # Non-authoritative SNMP engine: use cached estimates
+                if self.__timeline.has_key(securityEngineID):
+                    ( snmpEngineBoots,
+                      snmpEngineTime,
+                      latestReceivedEngineTime,
+                      latestUpdateTimestamp ) = self.__timeline[
+                        msgAuthoritativeEngineID
+                        ]
+                    # time passed since last talk with this SNMP engine
+                    idleTime = int(time.time())-latestUpdateTimestamp
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read timeline snmpEngineBoots %s snmpEngineTime %s for msgAuthoritativeEngineID %s, idle time %s secs' % (snmpEngineBoots, snmpEngineTime, msgAuthoritativeEngineID, idleTime))
+                else:
+                    raise error.ProtocolError('Peer SNMP engine info missing')
 
             msgAuthoritativeEngineBoots = securityParameters.getComponentByPosition(1)
             msgAuthoritativeEngineTime = securityParameters.getComponentByPosition(2)
@@ -598,7 +660,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             if msgAuthoritativeEngineID == snmpEngineID:
                 if snmpEngineBoots == 2147483647L or \
                    snmpEngineBoots != msgAuthoritativeEngineBoots or \
-                   abs(int(snmpEngineTime)-int(msgAuthoritativeEngineTime)) > 150:
+                   idleTime + abs(int(snmpEngineTime) - \
+                                  int(msgAuthoritativeEngineTime)) > 150:
                     usmStatsNotInTimeWindows, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsNotInTimeWindows')
                     usmStatsNotInTimeWindows.syntax = usmStatsNotInTimeWindows.syntax+1
                     raise error.StatusInformation(
@@ -620,14 +683,17 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     self.__timeline[msgAuthoritativeEngineID] = (
                         msgAuthoritativeEngineBoots,
                         msgAuthoritativeEngineTime,
-                        msgAuthoritativeEngineTime
+                        msgAuthoritativeEngineTime,
+                        int(time.time())
                         )
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: stored timeline msgAuthoritativeEngineBoots %s msgAuthoritativeEngineTime %s for msgAuthoritativeEngineID %s' % (msgAuthoritativeEngineBoots, msgAuthoritativeEngineTime, msgAuthoritativeEngineID))
                     
                 # 3.2.7b.2
                 if snmpEngineBoots == 2147483647L or \
                    msgAuthoritativeEngineBoots < snmpEngineBoots or \
                    msgAuthoritativeEngineBoots == snmpEngineBoots and \
-                   msgAuthoritativeEngineTime - snmpEngineTime < -150:
+                   idleTime + abs(int(snmpEngineTime) - \
+                                  int(msgAuthoritativeEngineTime)) > 150:
                     raise error.StatusInformation(
                         errorIndication = 'notInTimeWindow'
                         )
@@ -652,6 +718,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     securityParameters.getComponentByPosition(5),
                     encryptedPDU
                     )
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: PDU decyphered')
             except error.StatusInformation:
                 usmStatsDecryptionErrors, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsDecryptionErrors')
                 usmStatsDecryptionErrors.syntax = usmStatsDecryptionErrors.syntax+1
@@ -660,7 +727,6 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     oid=usmStatsDecryptionErrors.name,
                     val=usmStatsDecryptionErrors.syntax,
                     securityStateReference=securityStateReference,
-                    securityLevel=securityLevel,
                     contextEngineId=contextEngineId,
                     contextName=contextName,
                     maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
@@ -690,6 +756,8 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             usmUserPrivKeyLocalized=usmUserPrivKeyLocalized
             )
 
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: cached msgUserName %s info by securityStateReference %s' % (msgUserName, securityStateReference))
+        
         # Delayed to include details
         if not msgUserName and not securityEngineID:
             usmStatsUnknownUserNames, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownUserNames')
@@ -700,7 +768,6 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 val=usmStatsUnknownUserNames.syntax,
                 securityStateReference=securityStateReference,
                 securityEngineID=securityEngineID,
-                securityLevel=securityLevel,
                 contextEngineId=contextEngineId,
                 contextName=contextName,
                 maxSizeResponseScopedPDU=maxSizeResponseScopedPDU,
@@ -713,3 +780,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                  scopedPDU,
                  maxSizeResponseScopedPDU,
                  securityStateReference )
+
+    def __expireTimelineInfo(self):
+        if self.__timelineExpQueue.has_key(self.__expirationTimer):
+            for engineIdKey in self.__timelineExpQueue[self.__expirationTimer]:
+                del self.__timeline[engineIdKey]
+                debug.logger & debug.flagMP and debug.logger('__expireEnginesInfo: expiring %s' % (engineIdKey,))
+            del self.__timelineExpQueue[self.__expirationTimer]
+        self.__expirationTimer = self.__expirationTimer + 1
+        
+    def receiveTimerTick(self, snmpEngine, timeNow):
+        self.__expireTimelineInfo()
