@@ -4,6 +4,7 @@ from pysnmp.smi import exval, error
 from pysnmp.proto import rfc1902
 from pyasn1.type import constraint
 from pyasn1.error import ValueConstraintError
+from pysnmp import debug
 
 ( Integer, ObjectIdentifier, Null ) = mibBuilder.importSymbols("ASN1", "Integer", "ObjectIdentifier", "Null")
 
@@ -77,6 +78,20 @@ class ModuleIdentity(MibNode):
         self.revisions = args
         return self
 
+    def asn1Print(self):
+        return '\
+MODULE-IDENTITY\n\
+  LAST-UPDATED %s\n\
+  ORGANIZATION \"%s\"\n\
+  CONTACT-INFO \"%s\"\n\
+  DESCRIPTION \"%s\"\n\
+  %s\
+' % (self.getLastUpdated(),
+     self.getOrganization(),
+     self.getContactInfo(),
+     self.getDescription(),
+     string.join(map(lambda x: "REVISON \"%s\"\n" % x, self.getRevisions())))
+
 class ObjectIdentity(MibNode):
     def getStatus(self):
         return getattr(self, 'status', 'current')
@@ -94,6 +109,16 @@ class ObjectIdentity(MibNode):
         self.reference = v
         return self
 
+    def asn1Print(self):
+        return '\
+OBJECT-IDENTITY\n\
+  STATUS %s\n\
+  DESCRIPTION \"%s\"\n\
+  REFERENCE \"%s\"\
+' % (self.getStatus(),
+     self.getDescription(),
+     self.getReference())
+
 # definition for objects
 
 class NotificationType(MibNode):
@@ -118,7 +143,21 @@ class NotificationType(MibNode):
         self.revisions = args
         return self
 
-class MibIdentifier(MibNode): pass
+    def asn1Print(self):
+        return '\
+NOTIFICATION-TYPE\n\
+  OBJECTS { %s }\n\
+  STATUS %s\n\
+  DESCRIPTION \"%s\"\n\
+  %s\
+' % (reduce(lambda x,y: '%s, %s' % (x[1],y[1]), self.getObjects(), ("","")),
+     self.getStatus(),
+     self.getDescription(),
+     string.join(map(lambda x: "REVISON \"%s\"\n" % x, self.getRevisions())))
+
+class MibIdentifier(MibNode):
+    def asn1Print(self):
+        return 'OBJECT IDENTIFIER'
 
 class ObjectType(MibNode):
     maxAccess = None
@@ -161,6 +200,22 @@ class ObjectType(MibNode):
         self.reference = v
         return self
 
+    def asn1Print(self):
+        return '\
+OBJECT-TYPE\n\
+  SYNTAX %s\n\
+  UNITS \"%s\"\n\
+  MAX-ACCESS %s\n\
+  STATUS %s\n\
+  DESCRIPTION \"%s\"\n\
+  REFERENCE \"%s\"\
+' % (self.getSyntax().__class__.__name__,
+     self.getUnits(),
+     self.getMaxAccess(),
+     self.getStatus(),
+     self.getDescription(),
+     self.getReference())
+        
 class MibTree(ObjectType):
     branchVersionId = 0L    # increments on tree structure change XXX
     maxAccess = 'not-accessible'
@@ -203,7 +258,6 @@ class MibTree(ObjectType):
                 subName = name[:keyLen]
                 if self._vars.has_key(subName):
                     return self._vars[subName]
-        
         raise error.NoSuchObjectError(name=name, idx=idx)
 
     def getNode(self, name, idx=None):
@@ -242,7 +296,8 @@ class MibTree(ObjectType):
     
     def readTest(self, name, val, idx, (acFun, acCtx)):
         if name == self.name:
-            if self.maxAccess != 'readonly' and \
+            if acFun and \
+                   self.maxAccess != 'readonly' and \
                    self.maxAccess != 'readwrite' and \
                    self.maxAccess != 'readcreate' or \
                    acFun and acFun(name, idx, 'read', acCtx):
@@ -296,7 +351,8 @@ class MibTree(ObjectType):
     def writeTest(self, name, val, idx, (acFun, acCtx)):
         if name == self.name:
             # Make sure variable is writable
-            if self.maxAccess != 'readwrite' and \
+            if acFun and \
+                   self.maxAccess != 'readwrite' and \
                    self.maxAccess != 'readcreate' or \
                    acFun and acFun(name, idx, 'write', acCtx):
                 raise error.NotWritableError(idx=idx, name=name)
@@ -327,7 +383,8 @@ class MibScalar(MibTree):
         else:
             MibTree.readTest(self, name, val, idx, (acFun, acCtx))
         # If instance exists, check permissions
-        if self.maxAccess != 'readonly' and \
+        if acFun and \
+               self.maxAccess != 'readonly' and \
                self.maxAccess != 'readwrite' and \
                self.maxAccess != 'readcreate' or \
                acFun and acFun(name, idx, 'read', acCtx):
@@ -341,7 +398,8 @@ class MibScalar(MibTree):
         else:
             MibTree.writeTest(self, name, val, idx, (acFun, acCtx))
         # If instance exists, check permissions
-        if self.maxAccess != 'readwrite' and \
+        if acFun and \
+               self.maxAccess != 'readwrite' and \
                self.maxAccess != 'readcreate' or \
                acFun and acFun(name, idx, 'write', acCtx):
             raise error.NotWritableError(idx=idx, name=name)
@@ -369,12 +427,16 @@ class MibScalarInstance(MibTree):
     def readTest(self, name, val, idx, (acFun, acCtx)):
         if name != self.name:
             raise error.NoSuchObjectError(idx=idx, name=name)
-    
+
     def readGet(self, name, val, idx, (acFun, acCtx)):
         # Return current variable (name, value). This is the only API method
         # capable of returning anything!
         if name == self.name:
-            return self.name, self.syntax.clone()
+            debug.logger & debug.flagIns and debug.logger('readGet: %s=%s' % (self.name, self.syntax))
+            if hasattr(self.syntax, 'smiRead'):
+                return self.name, self.syntax.smiRead(name, val, idx)
+            else:
+                return self.name, self.syntax.clone()
         else:
             raise error.NoSuchObjectError(idx=idx, name=name)
     
@@ -383,7 +445,12 @@ class MibScalarInstance(MibTree):
     def writeTest(self, name, val, idx, (acFun, acCtx)):
         # Make sure write's allowed
         if name == self.name:
-            self.__newSyntax = self.syntax.clone(val)
+            if hasattr(self.syntax, 'smiWrite'):
+                self.__newSyntax = self.syntax.smiWrite(name, val, idx)
+            else:
+                self.__newSyntax = self.syntax.clone(val)
+            if hasattr(self.__newSyntax, 'smiRaisePendingError'):
+                self.__newSyntax.smiRaisePendingError()
         else:
             raise error.NoSuchObjectError(idx=idx, name=name)
 
@@ -392,6 +459,7 @@ class MibScalarInstance(MibTree):
         self.syntax, self.__newSyntax = self.__newSyntax, self.syntax
         
     def writeCleanup(self, name, val, idx, (acFun, acCtx)):
+        debug.logger & debug.flagIns and debug.logger('writeCleanup: %s=%s' % (name, val))
         # Drop previous value
         self.__newSyntax = None
     
@@ -404,12 +472,18 @@ class MibScalarInstance(MibTree):
     # Create operation
 
     def createTest(self, name, val, idx, (acFun, acCtx)):
-        if val is not None:
-            self.writeTest(name, val, idx, (acFun, acCtx))
+        if name == self.name:
+            if hasattr(self.syntax, 'smiCreate'):
+                self.__newSyntax = self.syntax.smiCreate(name, val, idx)
+            else:
+                self.__newSyntax = self.syntax.clone(val)
+        else:
+            raise error.NoSuchObjectError(idx=idx, name=name)
     def createCommit(self, name, val, idx, (acFun, acCtx)):
         if val is not None:
             self.writeCommit(name, val, idx, (acFun, acCtx))
     def createCleanup(self, name, val, idx, (acFun, acCtx)):
+        debug.logger & debug.flagIns and debug.logger('createCleanup: %s=%s' % (name, val))
         if val is not None:
             self.writeCleanup(name, val, idx, (acFun, acCtx))
     def createUndo(self, name, val, idx, (acFun, acCtx)):
@@ -418,7 +492,14 @@ class MibScalarInstance(MibTree):
 
     # Destroy operation
 
-    def destroyTest(self, name, val, idx, (acFun, acCtx)): pass
+    def destroyTest(self, name, val, idx, (acFun, acCtx)):
+        if name == self.name:
+            if hasattr(self.syntax, 'smiDestroy'):
+                self.__newSyntax = self.syntax.smiDestoy(name, val)
+            else:
+                self.__newSyntax = self.syntax.clone(val)
+        else:
+            raise error.NoSuchObjectError(idx=idx, name=name)
     def destroyCommit(self, name, val, idx, (acFun, acCtx)): pass
     def destroyCleanup(self, name, val, idx, (acFun, acCtx)): pass
     def destroyUndo(self, name, val, idx, (acFun, acCtx)): pass
@@ -457,22 +538,21 @@ class MibTableColumn(MibScalar):
         # do not replace the old one
         if name == self.name:
             raise error.NoAccessError(idx=idx, name=name)
-        if self._vars.has_key(name):
-            return
-        if val is not None and \
+        if acFun and \
+               val is not None and \
                self.maxAccess != 'readcreate' or \
                acFun and acFun(name, idx, 'write', acCtx):
             raise error.NoCreationError(idx=idx, name=name)
-        if not self.__createdInstances.has_key(name):
-            self.__createdInstances[name] = self.protoInstance(
-                self.name, name[len(self.name):], self.syntax.clone()
-                )
-        try:
-            self.__createdInstances[name].createTest(
-                name, val, idx, (acFun, acCtx)
-                )
-        except (error.RowCreationWanted, error.RowDestructionWanted):
-            pass
+        # Create instances if either is does not yet exist (row creation)
+        # or a value is passed (multiple OIDs in SET PDU)
+        if val is None and self.__createdInstances.has_key(name):
+            return
+        self.__createdInstances[name] = self.protoInstance(
+            self.name, name[len(self.name):], self.syntax.clone()
+            )
+        self.__createdInstances[name].createTest(
+            name, val, idx, (acFun, acCtx)
+            )
             
     def createCommit(self, name, val, idx, (acFun, acCtx)):
         # Commit new instance value
@@ -491,7 +571,7 @@ class MibTableColumn(MibScalar):
         # Drop previous column instance
         if self.__createdInstances.has_key(name):
             if self.__createdInstances[name] is not None:
-                self.__createdInstances[name].cleanupCleanup(
+                self.__createdInstances[name].createCleanup(
                     name, val, idx, (acFun, acCtx)
                     )
             del self.__createdInstances[name]
@@ -515,14 +595,16 @@ class MibTableColumn(MibScalar):
         # Make sure destruction is allowed
         if name == self.name:
             raise error.NoAccessError(idx=idx, name=name)        
-        if self._vars.has_key(name):
-            if val is not None and \
-                   self.maxAccess != 'readcreate' or \
-                   acFun and acFun(name, idx, 'write', cbCtx):
-                raise error.NoAccessError(idx=idx, name=name)
-            self._vars[name].destroyTest(
-                name, val, idx, (acFun, acCtx)
-                )
+        if not self._vars.has_key(name):
+            return
+        if acFun and \
+               val is not None and \
+               self.maxAccess != 'readcreate' or \
+               acFun and acFun(name, idx, 'write', acCtx):
+            raise error.NoAccessError(idx=idx, name=name)
+        self._vars[name].destroyTest(
+            name, val, idx, (acFun, acCtx)
+            )
 
     def destroyCommit(self, name, val, idx, (acFun, acCtx)):
         # Make a copy of column instance and take it off the tree
@@ -536,9 +618,10 @@ class MibTableColumn(MibScalar):
     def destroyCleanup(self, name, val, idx, (acFun, acCtx)):
         # Drop instance copy
         if self.__destroyedInstances.has_key(name):
-            self._destroyedInstances[name].destroyCleanup(
+            self.__destroyedInstances[name].destroyCleanup(
                 name, val, idx, (acFun, acCtx)
                 )
+            debug.logger & debug.flagIns and debug.logger('destroyCleanup: %s=%s' % (name, val))
             del self.__destroyedInstances[name]
             
     def destroyUndo(self, name, val, idx, (acFun, acCtx)):
@@ -561,12 +644,13 @@ class MibTableColumn(MibScalar):
                 )
         # ...otherwise proceed with creating new column
         except (error.NoSuchObjectError, error.RowCreationWanted):
-            self.__rowOpWanted[name] =  error.RowCreationWanted()
+            self.__rowOpWanted[name] = error.RowCreationWanted()
             self.createTest(name, val, idx, (acFun, acCtx))
         except error.RowDestructionWanted:
-            self.__rowOpWanted[name] =  error.RowDestructionWanted()
+            self.__rowOpWanted[name] = error.RowDestructionWanted()
             self.destroyTest(name, val, idx, (acFun, acCtx))
         if self.__rowOpWanted.has_key(name):
+            debug.logger & debug.flagIns and debug.logger('%s flagged by %s=%s' % (self.__rowOpWanted[name], name, val))
             raise self.__rowOpWanted[name]
 
     def __delegateWrite(self, subAction, name, val, idx, (acFun, acCtx)):
@@ -583,26 +667,33 @@ class MibTableColumn(MibScalar):
             getattr(self, 'destroy'+subAction)(
                 name, val, idx, (acFun, acCtx)
                 )
-        raise self.__rowOpWanted[name]
-
+        
     def writeCommit(self, name, val, idx, (acFun, acCtx)):
         self.__delegateWrite(
             'Commit', name, val, idx, (acFun, acCtx)
             )
+        if self.__rowOpWanted.has_key(name):
+            raise self.__rowOpWanted[name]
 
     def writeCleanup(self, name, val, idx, (acFun, acCtx)):
         self.__delegateWrite(
             'Cleanup', name, val, idx, (acFun, acCtx)
             )
         if self.__rowOpWanted.has_key(name):
+            e = self.__rowOpWanted[name]
             del self.__rowOpWanted[name]
+            debug.logger & debug.flagIns and debug.logger('%s dropped by %s=%s' % (e, name, val))
+            raise e
             
     def writeUndo(self, name, val, idx, (acFun, acCtx)):
         self.__delegateWrite(
             'Undo', name, val, idx, (acFun, acCtx)
             )
         if self.__rowOpWanted.has_key(name):
+            e = self.__rowOpWanted[name]
             del self.__rowOpWanted[name]
+            debug.logger & debug.flagIns and debug.logger('%s dropped by %s=%s' % (e, name, val))
+            raise e
 
 class MibTableRow(MibTree):
     """MIB table row (SMI 'Entry'). Manages a set of table columns.
@@ -613,22 +704,41 @@ class MibTableRow(MibTree):
         self.indexNames = ()
         self.augmentingRows = {}
 
-    # Table indices resolution
+    # Table indices resolution. Handle almost all possible rfc1902 types
+    # explicitly rather than by means of isSuperTypeOf() method because
+    # some subtypes may be implicitly tagged what renders base tag
+    # unavailable.
 
     __intValue = Integer()
+    __counter32Value = Counter32()
+    __uint32Value = Unsigned32()
+    __timeticksValue = TimeTicks()
+    __counter64Value = Counter64()
     __strValue = OctetString()
     __oidValue = ObjectIdentifier()
     __ipaddrValue = IpAddress()
+    __bitsValue = Bits()
 
     def setFromName(self, obj, value, impliedFlag=None):
-        if self.__intValue.isSuperTypeOf(obj):
+        if not value:
+            raise error.SmiError('Short OID for index %s' % repr(obj))
+        if self.__intValue.isSuperTypeOf(obj) or \
+               self.__uint32Value.isSuperTypeOf(obj) or \
+               self.__timeticksValue.isSuperTypeOf(obj) or \
+               self.__counter32Value.isSuperTypeOf(obj) or \
+               self.__counter64Value.isSuperTypeOf(obj):
             return obj.clone(value[0]), value[1:]
         elif self.__ipaddrValue.isSuperTypeOf(obj):
             return obj.clone(string.join(map(str, value[:4]), '.')), value[4:]
         elif self.__strValue.isSuperTypeOf(obj):
-            if impliedFlag:
+            # rfc1902, 7.7
+            if impliedFlag: 
                 s = reduce(lambda x,y: x+y, map(lambda x: chr(x), value))
                 return obj.clone(s), ()                
+            elif obj.isFixedLength():
+                len = obj.getFixedLength()
+                s = reduce(lambda x,y: x+y, map(lambda x: chr(x), value[:len]))
+                return obj.clone(s), value[len:]
             else:
                 s = reduce(lambda x,y: x+y,
                            map(lambda x: chr(x), value[1:value[0]+1]), '')
@@ -638,14 +748,26 @@ class MibTableRow(MibTree):
                 return obj.clone(value), ()
             else:
                 return obj.clone(value[1:value[0]+1]), value[value[0]+1:]
+        # rfc2578, 7.1
+        elif self.__bitsValue.isSuperTypeOf(obj):
+            s = reduce(
+                lambda x,y: x+y, map(lambda x: chr(x),value[1:value[0]+1]),''
+                )
+            return obj.clone(s), value[value[0]+1:]
         else:
             raise error.SmiError('Unknown value type for index %s' % repr(obj))
 
     def getAsName(self, obj, impliedFlag=None):
-        if self.__intValue.isSuperTypeOf(obj):
+        if self.__intValue.isSuperTypeOf(obj) or \
+               self.__uint32Value.isSuperTypeOf(obj) or \
+               self.__timeticksValue.isSuperTypeOf(obj) or \
+               self.__counter32Value.isSuperTypeOf(obj) or \
+               self.__counter64Value.isSuperTypeOf(obj):
             return (int(obj),)
+        elif self.__ipaddrValue.isSuperTypeOf(obj):
+            return tuple(map(ord, obj))
         elif self.__strValue.isSuperTypeOf(obj):
-            if impliedFlag:
+            if impliedFlag or obj.isFixedLength():
                 initial = ()
             else:
                 initial = (len(obj),)
@@ -657,8 +779,11 @@ class MibTableRow(MibTree):
                 return tuple(obj)
             else:
                 return (len(self.name),) + tuple(obj)
-        elif self.__ipaddrValue.isSuperTypeOf(obj):
-            return tuple(map(ord, obj))
+        # rfc2578, 7.1
+        elif self.__bitsValue.isSuperTypeOf(obj):
+            return reduce(
+                lambda x,y: x+(y,), map(lambda x: ord(x), obj),(len(obj),) 
+                )
         else:
             raise error.SmiError('Unknown value type for index %s' % repr(obj))
             
@@ -684,6 +809,7 @@ class MibTableRow(MibTree):
             return
         for modName, mibSym in self.augmentingRows.keys():
             mibObj, = mibBuilder.importSymbols(modName, mibSym)
+            debug.logger & debug.flagIns and debug.logger('announceManagementEvent %s to %s' % (action, mibObj))
             mibObj.receiveManagementEvent(
                 action, baseIndices, val, idx, (acFun, acCtx)
                 )
@@ -700,7 +826,9 @@ class MibTableRow(MibTree):
                 if name == mibObj.name:
                     newSuffix = newSuffix + self.getAsName(syntax, impliedFlag)
         if newSuffix:
-            self.__manageColumns(action, newSuffix, val, idx, (acFun, acCtx))
+            debug.logger & debug.flagIns and debug.logger('receiveManagementEvent %s for suffix %s' % (action, newSuffix))
+            self.__manageColumns(action, (), newSuffix, val, idx,
+                                 (acFun, acCtx))
 
     def registerAugmentions(self, *names):
         for modName, symName in names:
@@ -720,9 +848,13 @@ class MibTableRow(MibTree):
     def getIndexNames(self):
         return self.indexNames
                              
-    def __manageColumns(self, action, nameSuffix, val, idx, (acFun, acCtx)):
+    def __manageColumns(self, action, excludeName, nameSuffix,
+                        val, idx, (acFun, acCtx)):
         for name, var in self._vars.items():
+            if name == excludeName:
+                continue
             getattr(var, action)(name + nameSuffix, val, idx, (acFun, acCtx))
+            debug.logger & debug.flagIns and debug.logger('__manageColumns: action %s name %s suffix %s value %s' % (action, name, nameSuffix, val))
 
     def __delegate(self, subAction, name, val, idx, (acFun, acCtx)):
         # Relay operation request to column, expect row operation request.
@@ -732,16 +864,16 @@ class MibTableRow(MibTree):
                 )
         except error.RowCreationWanted, why:
             self.__manageColumns(
-                'create'+subAction, name[len(self.name)+1:],
-                None, idx, (acFun, acCtx)
+                'create'+subAction, name[:len(self.name)+1],
+                name[len(self.name)+1:], None, idx, (acFun, acCtx)
                 )
             self.announceManagementEvent(
                 'create'+subAction, name, None, idx, (acFun, acCtx)
                 )
         except error.RowDestructionWanted, why:
             self.__manageColumns(
-                'destroy'+subAction, name[len(self.name)+1:],
-                None, idx, (acFun, acCtx)
+                'destroy'+subAction, name[:len(self.name)+1],
+                name[len(self.name)+1:], None, idx, (acFun, acCtx)
                 )
             self.announceManagementEvent(
                 'destroy'+subAction, name, None, idx, (acFun,acCtx)
